# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Residual Attention network, implemented in Gluon."""

from mxnet.gluon import nn
from mxnet.gluon.model_zoo.vision.resnet import BottleneckV2
from ..nn.basic_blocks import FrBase


__all__ = ["AttentionNet", "AttentionNetFace",
           "get_attention_net", "get_attention_face",
           "attention_net56", "attention_net92", "attention_net128",
           "attention_net164", "attention_net236", "attention_net452"
           ]


class AttentionBlock(nn.HybridBlock):
    def __init__(self, channels, out_size, stage, p=1, t=2, r=1, **kwargs):
        r"""Residual Attention Block from
        `"Residual Attention Network for Image Classification"
        <https://arxiv.org/abs/1704.06904>`_ paper.

        Parameters
        ----------
        :param channels: int. Number of output channels.
        :param out_size: int. Size of the output feature map, now it only supports square shape.
        :param stage: int. Stage described in Figure 2.
        :param p: int. Number of pre-processing Residual Units before split into trunk branch and mask branch.
        :param t: int. Number of Residual Units in trunk branch.
        :param r: int. Number of Residual Units between adjacent pooling layer in the mask branch.
        :param kwargs:
        """
        super().__init__(**kwargs)
        with self.name_scope():
            self.pre = nn.HybridSequential()
            for i in range(p):
                self.pre.add(BottleneckV2(channels, 1, prefix='pre_%d_' % i))

            self.trunk_branch = nn.HybridSequential()
            for i in range(t):
                self.trunk_branch.add(BottleneckV2(channels, 1, prefix='trunk_%d_' % i))

            self.mask_branch = _MaskBlock(channels, r, out_size, stage, prefix='mask_')

            self.post = nn.HybridSequential()
            for i in range(p):
                self.post.add(BottleneckV2(channels, 1, prefix='post_%d_' % i))

    def hybrid_forward(self, F, x, *args, **kwargs):
        x = self.pre(x)
        mask = self.mask_branch(x)
        trunk = self.trunk_branch(x)
        out = (1 + mask) * trunk
        out = self.post(out)
        return out


class _UpSampleBlock(nn.HybridBlock):
    def __init__(self, out_size, **kwargs):
        super().__init__(**kwargs)
        self._size = out_size

    def hybrid_forward(self, F, x, *args, **kwargs):
        return F.contrib.BilinearResize2D(x, height=self._size, width=self._size)


class _MaskBlock(nn.HybridBlock):
    def __init__(self, channels, r, out_size, stage, **kwargs):
        super().__init__(**kwargs)
        self._stage = stage
        with self.name_scope():
            self._make_layers(channels, r, stage, out_size)

    def _make_layers(self, channels, r, stage, out_size):
        if stage <= 1:
            self.down_sample_1 = nn.MaxPool2D(3, 2, 1)
            self.down_res_unit_1 = nn.HybridSequential()
            for i in range(r):
                self.down_res_unit_1.add(BottleneckV2(channels, 1, prefix="down_res1_%d_" % i))
            self.skip_connection_1 = BottleneckV2(channels, 1)

            self.up_res_unit_1 = nn.HybridSequential()
            for i in range(r):
                self.up_res_unit_1.add(BottleneckV2(channels, 1, prefix="up_res1_%d_" % i))
            self.up_sample_1 = _UpSampleBlock(out_size)
            out_size = out_size // 2

        if stage <= 2:
            self.down_sample_2 = nn.MaxPool2D(3, 2, 1)
            self.down_res_unit_2 = nn.HybridSequential()
            for i in range(r):
                self.down_res_unit_2.add(BottleneckV2(channels, 1, prefix="down_res2_%d_" % i))
            self.skip_connection_2 = BottleneckV2(channels, 1)

            self.up_res_unit_2 = nn.HybridSequential()
            for i in range(r):
                self.up_res_unit_2.add(BottleneckV2(channels, 1, prefix="up_res2_%d_" % i))
            self.up_sample_2 = _UpSampleBlock(out_size)
            out_size = out_size // 2

        if stage <= 3:
            self.down_sample_3 = nn.MaxPool2D(3, 2, 1)
            self.down_res_unit_3 = nn.HybridSequential()
            for i in range(r):
                self.down_res_unit_3.add(BottleneckV2(channels, 1, prefix="down_res3_%d_" % i))

            self.up_res_unit_3 = nn.HybridSequential()
            for i in range(r):
                self.up_res_unit_3.add(BottleneckV2(channels, 1, prefix="up_res3_%d_" % i))
            self.up_sample_3 = _UpSampleBlock(out_size)

        self.output = nn.HybridSequential()
        self.output.add(nn.BatchNorm(),
                        nn.Activation('relu'),
                        nn.Conv2D(channels, kernel_size=1, strides=1, use_bias=False),
                        nn.BatchNorm(),
                        nn.Activation('relu'),
                        nn.Conv2D(channels, kernel_size=1, strides=1, use_bias=False),
                        nn.Activation('sigmoid')
                        )

    def hybrid_forward(self, F, x, *args, **kwargs):
        if self._stage <= 1:
            x_down1 = self.down_sample_1(x)
            x_down1 = self.down_res_unit_1(x_down1)
            residual_1 = self.skip_connection_1(x_down1)
        else:
            x_down1 = x
            residual_1 = 0

        if self._stage <= 2:
            x_down2 = self.down_sample_2(x_down1)
            x_down2 = self.down_res_unit_2(x_down2)
            residual_2 = self.skip_connection_2(x_down2)
        else:
            x_down2 = x
            residual_2 = 0

        if self._stage <= 3:
            x_down3 = self.down_sample_3(x_down2)
            x_down3 = self.down_res_unit_3(x_down3)

            x_up3 = self.up_res_unit_3(x_down3)
            x_up3 = self.up_sample_3(x_up3)
        else:
            raise ValueError("param stage should be a number not larger than 3!")

        if self._stage <= 2:
            x_up2 = x_up3 + residual_2
            x_up2 = self.up_res_unit_2(x_up2)
            x_up2 = self.up_sample_2(x_up2)
        else:
            x_up2 = x_up3

        if self._stage <= 1:
            x_up1 = x_up2 + residual_1
            x_up1 = self.up_res_unit_1(x_up1)
            x_up1 = self.up_sample_1(x_up1)
        else:
            x_up1 = x_up2
        out = self.output(x_up1)
        return out


class AttentionNet(nn.HybridBlock):
    r"""AttentionNet Model from
    `"Residual Attention Network for Image Classification"
    <https://arxiv.org/abs/1704.06904>`_ paper.

    Parameters
    ----------
    :param classes: int. Number of classification classes.
    :param modules: list. The number of Attention Module in each stage.
    :param p: int. Number of pre-processing Residual Units before split into trunk branch and mask branch.
    :param t: int. Number of Residual Units in trunk branch.
    :param r: int. Number of Residual Units between adjacent pooling layer in the mask branch.
    :param kwargs:

    """

    def __init__(self, classes, modules, p, t, r, **kwargs):
        super().__init__(**kwargs)
        assert len(modules) == 3
        with self.name_scope():
            self.features = nn.HybridSequential()
            # 112x112
            self.features.add(nn.Conv2D(64, 3, 2, 1, use_bias=False))
            self.features.add(nn.BatchNorm())
            self.features.add(nn.Activation('relu'))

            # 56x56
            self.features.add(nn.MaxPool2D(3, 2, 1))
            self.features.add(BottleneckV2(256, 1, True, 64))
            for _ in range(modules[0]):
                self.features.add(AttentionBlock(256, 56, 1, p, t, r))

            # 28x28
            self.features.add(BottleneckV2(512, 2, True, 256))
            for _ in range(modules[1]):
                self.features.add(AttentionBlock(512, 28, 2, p, t, r))

            # 14x14
            self.features.add(BottleneckV2(1024, 2, True, 512))
            for _ in range(modules[2]):
                self.features.add(AttentionBlock(1024, 14, 3, p, t, r))

            # 7x7
            self.features.add(BottleneckV2(2048, 2, True, 1024),
                              BottleneckV2(2048, 1),
                              BottleneckV2(2048, 1))

            # 2048
            self.features.add(nn.BatchNorm(),
                              nn.Activation('relu'),
                              nn.GlobalAvgPool2D(),
                              nn.Flatten())

            # classes
            self.output = nn.Dense(classes)

    def hybrid_forward(self, F, x, *args, **kwargs):
        x = self.features(x)
        x = self.output(x)
        return x


class AttentionNetFace(FrBase):
    r"""
    AttentionNet Model for input 112x112.

    Parameters
    ----------
    :param classes: int. Number of classification classes.
    :param kwargs:

    """

    def __init__(self, classes, modules, p, t, r,
                 weight_norm=False, feature_norm=False, embedding_size=512,
                 need_cls_layer=True, **kwargs):
        super().__init__(classes, embedding_size, weight_norm, feature_norm, need_cls_layer, **kwargs)
        assert len(modules) == 3
        with self.name_scope():
            self.features = nn.HybridSequential()
            # 112x112
            self.features.add(nn.Conv2D(64, 3, 1, 1, use_bias=False))
            self.features.add(nn.BatchNorm())
            self.features.add(nn.Activation('relu'))

            # 56x56
            self.features.add(BottleneckV2(256, 2, True, 64))
            for _ in range(modules[0]):
                self.features.add(AttentionBlock(256, 56, 1, p, t, r))

            # 28x28
            self.features.add(BottleneckV2(512, 2, True, 256))
            for _ in range(modules[1]):
                self.features.add(AttentionBlock(512, 28, 2, p, t, r))

            # 14x14
            self.features.add(BottleneckV2(1024, 2, True, 512))
            for _ in range(modules[2]):
                self.features.add(AttentionBlock(1024, 14, 3, p, t, r))

            # 8x8
            self.features.add(BottleneckV2(2048, 2, True, 1024),
                              BottleneckV2(2048, 1),
                              BottleneckV2(2048, 1))

            # 2048
            self.features.add(nn.BatchNorm(),
                              nn.Activation('relu'),
                              nn.GlobalAvgPool2D(),
                              nn.Flatten())
            # embedding
            self.features.add(nn.Dense(embedding_size, use_bias=False),
                              nn.BatchNorm(scale=False, center=False),
                              nn.PReLU())


# Specification ([p, t, r], [stage1, stage2, stage3])
# The hyper-parameters are based on paper section 4.1. The number of layers can be calculated by 36m+20
# where m is the number of Attention Module in each stage when `p, t, r = 1, 2, 1`.
attention_net_spec = {56: ([1, 2, 1], [1, 1, 1]),
                      92: ([1, 2, 1], [1, 2, 3]),
                      128: ([1, 2, 1], [3, 3, 3]),
                      164: ([1, 2, 1], [4, 4, 4]),
                      236: ([1, 2, 1], [6, 6, 6]),
                      452: ([2, 4, 3], [6, 6, 6])}


# Constructor
def get_attention_net(classes, num_layers, **kwargs):
    r"""AttentionNet Model from
    `"Residual Attention Network for Image Classification"
    <https://arxiv.org/abs/1704.06904>`_ paper.

    Parameters
    ----------
    classes : int,
        Number of classification classes.
    num_layers : int
        Numbers of layers. Options are 56, 92, 128, 164, 236, 452.
    """
    assert num_layers in attention_net_spec, \
        "Invalid number of layers: %d. Options are %s" % (
            num_layers, str(attention_net_spec.keys()))
    ptr, modules = attention_net_spec[num_layers]
    assert len(ptr) == len(modules) == 3
    p, t, r = ptr
    net = AttentionNet(classes, modules, p, t, r, **kwargs)
    return net


def get_attention_face(classes, num_layers, embedding_size, **kwargs):
    r"""AttentionNet Model for 112x112 face images from
    `"Residual Attention Network for Image Classification"
    <https://arxiv.org/abs/1704.06904>`_ paper.

    Parameters
    ----------
    classes : int,
        Number of classification classes.
    num_layers : int
        Numbers of layers. Options are 56, 92, 128, 164, 236, 452.
    embedding_size: int
        Feature dimensions of the embedding layers.
    """
    assert num_layers in attention_net_spec, \
        "Invalid number of layers: %d. Options are %s" % (
            num_layers, str(attention_net_spec.keys()))
    ptr, modules = attention_net_spec[num_layers]
    assert len(ptr) == len(modules) == 3
    p, t, r = ptr
    net = AttentionNetFace(classes, modules, p, t, r, embedding_size=embedding_size, **kwargs)
    return net


def attention_net56(classes, **kwargs):
    r"""AttentionNet 56 Model from
       `"Residual Attention Network for Image Classification"
       <https://arxiv.org/abs/1704.06904>`_ paper.

       Parameters
       ----------
       classes : int,
           Number of classification classes.
       """
    return get_attention_net(classes, 56, **kwargs)


def attention_net92(classes, **kwargs):
    r"""AttentionNet 92 Model from
       `"Residual Attention Network for Image Classification"
       <https://arxiv.org/abs/1704.06904>`_ paper.

       Parameters
       ----------
       classes : int,
           Number of classification classes.
       """
    return get_attention_net(classes, 92, **kwargs)


def attention_net128(classes, **kwargs):
    r"""AttentionNet 128 Model from
       `"Residual Attention Network for Image Classification"
       <https://arxiv.org/abs/1704.06904>`_ paper.

       Parameters
       ----------
       classes : int,
           Number of classification classes.
       """
    return get_attention_net(classes, 128, **kwargs)


def attention_net164(classes, **kwargs):
    r"""AttentionNet 164 Model from
       `"Residual Attention Network for Image Classification"
       <https://arxiv.org/abs/1704.06904>`_ paper.

       Parameters
       ----------
       classes : int,
           Number of classification classes.
       """
    return get_attention_net(classes, 164, **kwargs)


def attention_net236(classes, **kwargs):
    r"""AttentionNet 236 Model from
       `"Residual Attention Network for Image Classification"
       <https://arxiv.org/abs/1704.06904>`_ paper.

       Parameters
       ----------
       classes : int,
           Number of classification classes.
       """
    return get_attention_net(classes, 236, **kwargs)


def attention_net452(classes, **kwargs):
    r"""AttentionNet 452 Model from
       `"Residual Attention Network for Image Classification"
       <https://arxiv.org/abs/1704.06904>`_ paper.

       Parameters
       ----------
       classes : int,
           Number of classification classes.
       """
    return get_attention_net(classes, 452, **kwargs)



# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Custom losses"""
import math
import numpy as np
from mxnet import nd, init
from mxnet.gluon.loss import Loss, SoftmaxCrossEntropyLoss

__all__ = ["get_loss", "SoftmaxCrossEntropyLoss", "ArcLoss", "TripletLoss", "RingLoss",
           "CosLoss", "L2Softmax", "ASoftmax", "CenterLoss", "ContrastiveLoss", "LGMLoss",
           "MPSLoss", "GitLoss", "COCOLoss"]
numeric_types = (float, int, np.generic)


def _apply_weighting(F, loss, weight=None, sample_weight=None):
    """Apply weighting to loss.

    Parameters
    ----------
    loss : Symbol
        The loss to be weighted.
    weight : float or None
        Global scalar weight for loss.
    sample_weight : Symbol or None
        Per sample weighting. Must be broadcastable to
        the same shape as loss. For example, if loss has
        shape (64, 10) and you want to weight each sample
        in the batch separately, `sample_weight` should have
        shape (64, 1).

    Returns
    -------
    loss : Symbol
        Weighted loss
    """
    if sample_weight is not None:
        loss = F.broadcast_mul(loss, sample_weight)

    if weight is not None:
        assert isinstance(weight, numeric_types), "weight must be a number"
        loss = loss * weight

    return loss


def _reshape_like(F, x, y):
    """Reshapes x to the same shape as y."""
    return x.reshape(y.shape) if F is nd.ndarray else F.reshape_like(x, y)


class L2Softmax(SoftmaxCrossEntropyLoss):
    r"""L2Softmax from
    `"L2-constrained Softmax Loss for Discriminative Face Verification"
    <https://arxiv.org/abs/1703.09507>`_ paper.

    Parameters
    ----------
    classes: int.
        Number of classes.
    alpha: float.
        The scaling parameter, a hypersphere with small alpha
        will limit surface area for embedding features.
    p: float, default is 0.9.
        The expected average softmax probability for correctly
        classifying a feature.
    from_normx: bool, default is False.
         Whether input has already been normalized.

    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.
    """

    def __init__(self, classes, alpha, p=0.9, from_normx=False,
                 axis=-1, sparse_label=True, weight=None, batch_axis=0, **kwargs):
        super().__init__(axis=axis, sparse_label=sparse_label, weight=weight, batch_axis=batch_axis, **kwargs)
        alpha_low = math.log(p * (classes - 2) / (1 - p))
        assert alpha > alpha_low, "For given probability of p={}, alpha should higher than {}.".format(p, alpha_low)
        self.alpha = alpha
        self._from_normx = from_normx

    def hybrid_forward(self, F, x, label, sample_weight=None):
        if not self._from_normx:
            x = F.L2Normalization(x, mode='instance', name='fc1n')
        fc7 = x * self.alpha
        return super().hybrid_forward(F, pred=fc7, label=label, sample_weight=sample_weight)


class CosLoss(SoftmaxCrossEntropyLoss):
    r"""CosLoss from
       `"CosFace: Large Margin Cosine Loss for Deep Face Recognition"
       <https://arxiv.org/abs/1801.09414>`_ paper.

       It is also AM-Softmax from
       `"Additive Margin Softmax for Face Verification"
       <https://arxiv.org/abs/1801.05599>`_ paper.

    Parameters
    ----------
    classes: int.
        Number of classes.
    m: float, default 0.4
        Margin parameter for loss.
    s: int, default 64
        Scale parameter for loss.


    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.
    """

    def __init__(self, classes, m, s, dtype="float32", **kwargs):
        super().__init__(**kwargs)
        self._classes = classes
        self._scale = s
        self._margin = m
        self._dtype = dtype

    def hybrid_forward(self, F, x, label, sample_weight=None):
        if self._sparse_label:
            one_hot_label = F.one_hot(label, depth=self._classes, on_value=1.0, off_value=0.0, dtype=self._dtype)
        else:
            one_hot_label = label

        body = one_hot_label * self._margin
        fc7 = (x - body) * self._scale

        return super().hybrid_forward(F, pred=fc7, label=label, sample_weight=sample_weight)


class ArcLoss(SoftmaxCrossEntropyLoss):
    r"""ArcLoss from
    `"ArcFace: Additive Angular Margin Loss for Deep Face Recognition"
    <https://arxiv.org/abs/1801.07698>`_ paper.

    Parameters
    ----------
    classes: int.
        Number of classes.
    m: float.
        Margin parameter for loss.
    s: int.
        Scale parameter for loss.

    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.
    """

    def __init__(self, classes, m=0.5, s=64, easy_margin=True, dtype="float32", **kwargs):
        super().__init__(**kwargs)
        assert s > 0.
        assert 0 <= m < (math.pi / 2)
        self.s = s
        self.m = m
        self.cos_m = math.cos(m)
        self.sin_m = math.sin(m)
        self.mm = math.sin(math.pi - m) * m
        self.threshold = math.cos(math.pi - m)
        self._classes = classes
        self.easy_margin = easy_margin
        self._dtype = dtype

    def hybrid_forward(self, F, pred, label, sample_weight=None, *args, **kwargs):
        cos_t = F.pick(pred, label, axis=1)  # cos(theta_yi)
        if self.easy_margin:
            cond = F.Activation(data=cos_t, act_type='relu')
        else:
            cond_v = cos_t - self.threshold
            cond = F.Activation(data=cond_v, act_type='relu')

        # sin_t = F.sqrt(1.0 - cos_t * cos_t)  # sin(theta)
        # new_zy = cos_t * self.cos_m - sin_t * self.sin_m  # cos(theta_yi + m)

        new_zy = F.cos(F.arccos(cos_t) + self.m)  # cos(theta_yi + m)
        if self.easy_margin:
            zy_keep = cos_t
        else:
            zy_keep = cos_t - self.mm  # (cos(theta_yi) - sin(pi - m)*m)
        new_zy = F.where(cond, new_zy, zy_keep)
        diff = new_zy - cos_t  # cos(theta_yi + m) - cos(theta_yi)
        diff = F.expand_dims(diff, 1)  # shape=(b, 1)
        gt_one_hot = F.one_hot(label, depth=self._classes, on_value=1.0, off_value=0.0, dtype=self._dtype)
        body = F.broadcast_mul(gt_one_hot, diff)
        pred = pred + body
        pred = pred * self.s

        return super().hybrid_forward(F, pred=pred, label=label, sample_weight=sample_weight)


class TripletLoss(Loss):
    r"""Calculates triplet loss given three input tensors and a positive margin.
    Triplet loss measures the relative similarity between prediction, a positive
    example and a negative example:

    .. math::
        L = \sum_i \max(\Vert {pred}_i - {pos_i} \Vert_2^2 -
                        \Vert {pred}_i - {neg_i} \Vert_2^2 + {margin}, 0)

    `pred`, `positive` and `negative` can have arbitrary shape as long as they
    have the same number of elements.

    Parameters
    ----------
    margin: float
        Margin of separation between correct and incorrect pair.
    weight: float or None
        Global scalar weight for loss.
    batch_axis: int, default 0
        The axis that represents mini-batch.


    Inputs:
        - **pred**: prediction tensor with arbitrary shape
        - **positive**: positive example tensor with arbitrary shape. Must have
          the same size as pred.
        - **negative**: negative example tensor with arbitrary shape Must have
          the same size as pred.

    Outputs:
        - **loss**: loss tensor with shape (batch_size,).
    """

    def __init__(self, margin=1, weight=None, batch_axis=0, **kwargs):
        super(TripletLoss, self).__init__(weight, batch_axis, **kwargs)
        self._margin = margin

    def hybrid_forward(self, F, pred, positive, negative):
        positive = _reshape_like(F, positive, pred)
        negative = _reshape_like(F, negative, pred)
        loss = F.sum(F.square(pred - positive) - F.square(pred - negative),
                     axis=self._batch_axis, exclude=True)
        loss = F.relu(loss + self._margin)
        return _apply_weighting(F, loss, self._weight, None)


class ContrastiveLoss(Loss):
    r"""Computes the contrastive loss.
    This loss encourages the embedding to be close to each other for
    the samples of the same label and the embedding to be far apart at least
    by the margin constant for the samples of different labels.
    See: http://yann.lecun.com/exdb/publis/pdf/hadsell-chopra-lecun-06.pdf
    Parameters
    ----------
    margin: float, default is 1.
        Margin term in the loss definition.

    Inputs:

        - **anchor**: prediction tensor. Embeddings should be l2 normalized.
        - **positive**: positive example tensor with arbitrary shape. Must have
          the same size as anchor. Embeddings should be l2 normalized.
        - **labels**: array with shape (batch_size,) of
          binary labels indicating positive vs negative pair.

    Outputs:
        - **loss**:  loss tensor with shape (batch_size,).Dimensions other than
          batch_axis are averaged out.
      """

    def __init__(self, margin=1, weight=None, batch_axis=0, **kwargs):
        super().__init__(weight, batch_axis, **kwargs)
        self._margin = margin

    def hybrid_forward(self, F, anchor, positive, labels):
        positive = _reshape_like(F, positive, anchor)
        dists = F.norm(F.square(anchor - positive), axis=1)
        loss = labels * F.square(dists) + (1 - labels) * F.square(F.maximum(self._margin - dists, 0))
        return _apply_weighting(F, loss, self._weight, None)


class RingLoss(SoftmaxCrossEntropyLoss):
    """Computes the Ring Loss from
    `"Ring loss: Convex Feature Normalization for Face Recognition"
    <https://arxiv.org/abs/1803.00130>`_paper.

    .. math::
        p = \softmax({pred})

        L_SM = -\sum_i \log p_{i,{label}_i}

        L_R = \frac{\lambda}{2m} \sum_{i=1}^{m} (\Vert \mathcal{F}({x}_i)\Vert_2 - R )^2

    Parameters
    ----------
    lamda: float
        The loss weight enforcing a trade-off between the softmax loss and ring loss.

    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.

    """

    def __init__(self, lamda, weight_initializer=None, dtype='float32', **kwargs):
        super().__init__(**kwargs)

        self._lamda = lamda
        self.R = self.params.get('R', shape=(1,), init=weight_initializer,
                                 dtype=dtype, allow_deferred_init=True)

    def hybrid_forward(self, F, pred, label, embedding, R, sample_weight=None):
        # RingLoss
        emb_norm = F.norm(embedding, axis=1)
        loss_r = F.square(F.broadcast_sub(emb_norm, R)) * 0.5
        loss_r = _apply_weighting(F, loss_r, self._weight, sample_weight)

        # Softmax
        loss_sm = super().hybrid_forward(F, pred, label, sample_weight)

        return loss_sm + self._lamda * loss_r


class ASoftmax(SoftmaxCrossEntropyLoss):
    r"""ASoftmax from
    `"SphereFace: Deep Hypersphere Embedding for Face Recognition"
    <https://arxiv.org/pdf/1704.08063.pdf>`_ paper.
    input(weight, x) has already been normalized

    Parameters
    ----------
    classes: int.
        Number of classes.
    m: float.
        Margin parameter for loss.
    s: int.
        Scale parameter for loss.

    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.
    """

    def __init__(self, classes, m, s, phiflag=True, dtype="float32", **kwargs):
        super().__init__(**kwargs)
        self._classes = classes
        self._scale = s
        self._margin = m
        self._phiflag = phiflag
        self._dtype = dtype
        self.it = 0
        self.LambdaMin = 5.0
        self.LambdaMax = 1500.0
        self.lamb = 1500.0
        self.mlambda = [
            lambda x: x ** 0,
            lambda x: x ** 1,
            lambda x: 2 * x ** 2 - 1,
            lambda x: 4 * x ** 3 - 3 * x,
            lambda x: 8 * x ** 4 - 8 * x ** 2 + 1,
            lambda x: 16 * x ** 5 - 20 * x ** 3 + 5 * x
        ]

    @staticmethod
    def _myphi(x, m):
        x = x * m
        return 1 - x ** 2 / math.factorial(2) + x ** 4 / math.factorial(4) - x ** 6 / math.factorial(6) + \
               x ** 8 / math.factorial(8) - x ** 9 / math.factorial(9)

    def hybrid_forward(self, F, x, label, sample_weight=None):
        cos_theta = F.clip(x, -1, 1)

        if self._phiflag:
            cos_m_theta = self.mlambda[int(self._margin)](cos_theta)
            theta = cos_theta.arccos()
            k = (self._margin * theta / math.pi).floor()
            n_one = k * 0.0 - 1
            phi_theta = (n_one ** k) * cos_m_theta - 2 * k
        else:
            theta = cos_theta.arccos()
            phi_theta = self._myphi(theta, self._margin)
            phi_theta = phi_theta.clip(-1 * self._margin, 1)

        if self._sparse_label:
            one_hot_label = F.one_hot(label, depth=self._classes, on_value=1.0, off_value=0.0, dtype=self._dtype)
        else:
            one_hot_label = label

        self.it += 1
        self.lamb = max(self.LambdaMin, self.LambdaMax / (1 + 0.1 * self.it))
        diff = (phi_theta - x) * 1.0 / (1 + self.lamb)

        body = one_hot_label * diff
        fc7 = (x + body) * self._scale

        return super().hybrid_forward(F, pred=fc7, label=label, sample_weight=sample_weight)


class CenterLoss(SoftmaxCrossEntropyLoss):
    """Computes the Center Loss from
    `"A Discriminative Feature Learning Approach for Deep Face Recognition"
    <http://ydwen.github.io/papers/WenECCV16.pdf>`_paper.
    Implementation is refer to
    "https://github.com/ShownX/mxnet-center-loss/blob/master/center_loss.py"

    Parameters
    ----------
    classes: int.
        Number of classes.

    lamda: float
        The loss weight enforcing a trade-off between the softmax loss and center loss.

    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.

    """

    def __init__(self, classes, embedding_size, lamda, weight_initializer=init.Xavier(magnitude=2.24),
                 dtype='float32', **kwargs):
        super().__init__(**kwargs)
        self._lamda = lamda
        self._classes = classes
        self._dtype = dtype
        self.centers = self.params.get('centers', shape=(classes, embedding_size), init=weight_initializer,
                                       dtype=dtype, allow_deferred_init=True)

    def hybrid_forward(self, F, x, label, embeddings, centers, sample_weight=None):
        # loss center
        centers_count = F.take(F.sum(F.one_hot(label, depth=self._classes, dtype=self._dtype), axis=0), label)
        centers_selected = F.take(centers, label)
        loss_c = self._lamda * 0.5 * F.sum(F.square(embeddings - centers_selected), 1) / centers_count

        # Softmax
        loss_sm = super().hybrid_forward(F, x, label, sample_weight)
        return loss_sm + loss_c


class LGMLoss(Loss):
    """LGM Loss from
    `"Rethinking Feature Distribution for Loss Functions in Image Classification"
    <https://arxiv.org/abs/1803.02988>`_paper.
    Implementation is refer to
    https://github.com/LeeJuly30/L-GM-Loss-For-Gluon/blob/master/L_GM.py

    Parameters
    ----------
    num_classes: int.
        The num of classes.
    embedding_size: int.
        The size of embedding feature.
    alpha: float.
        A non-negative parameter controlling the size of the expected margin between
        two classes on the training set.
    lamda: float.
        A non-negative weighting coefficient.
    lr_mult: float.
        Var updating need a relatively low learning rate compared to the overall learning rate.
    """

    def __init__(self, num_classes, embedding_size, alpha, lamda, lr_mult, dtype="float32", **kwargs):
        super().__init__(weight=None, batch_axis=0, **kwargs)
        self._num_class = num_classes
        self._feature_dim = embedding_size
        self._alpha = alpha
        self._lamda = lamda
        self._dtype = dtype
        self.mean = self.params.get('mean', shape=(num_classes, embedding_size), init=init.Xavier())
        self.var = self.params.get('var', shape=(num_classes, embedding_size), init=init.Constant(1), lr_mult=lr_mult)

    def _classification_probability(self, F, x, label, mean, var):
        reshape_var = F.reshape(var, (-1, 1, self._feature_dim))
        reshape_mean = F.reshape(mean, (-1, 1, self._feature_dim))
        x = F.expand_dims(x, 0)
        x = F.broadcast_minus(x, reshape_mean)
        d_z = F.elemwise_mul(F.broadcast_div(x, (reshape_var + 1e-8)), x)
        d_z = F.transpose(F.sum(d_z, axis=2) / 2)

        mask = F.one_hot(label, self._num_class, dtype=self._dtype) * self._alpha + 1
        margin_d_z = d_z * mask
        probability = F.broadcast_div(F.exp(-margin_d_z), (F.sqrt(F.prod(var, 1)) + 1e-8))
        return probability, d_z

    def hybrid_forward(self, F, x, label, mean, var):
        probability, m_distance = self._classification_probability(F, x, label, mean, var)

        # classification loss
        class_probability = F.pick(probability, label, axis=1)
        loss_cls = -F.log(class_probability / (F.sum(probability, 1) + 1e-8) + 1e-8)

        # likehood loss
        loss_lkd = F.pick(m_distance, label, axis=1)
        l_gm_loss = loss_cls + self._lamda * loss_lkd
        return l_gm_loss, probability


class RangeLoss(Loss):
    """Range Loss from
    `"Range Loss for Deep Face Recognition with Long-tail"
    <https://arxiv.org/abs/1611.08976>`_paper.
    Implementation is refer to
    https://github.com/LeeJuly30/RangeLoss-For-Gluno/blob/master/RangeLossForGluon.py

    L = L_SM + lamda*(alpha*L_Rintra + beta*L_Rinter)

    Parameters
    ----------
    alpha: float.
        Weight of L_Rintra.
    beta: float.
        Weight of L_Rinter.
    top_k: int.
        Compute the first k-largest range within each class, k=2 is recommended in paper.
    num_class: int.

    num_in_class: int.

    margin: float.
        Denotes a super parameter as the max optimization margin that will exclude DCenter greater than
        this margin from the computation of the inter loss.

    TODO: this loss has not been tested and made to be hybridize.
    """

    def __init__(self, alpha, beta, top_k, num_class, num_in_class, feature_dim, margin, **kwargs):
        super(RangeLoss, self).__init__(weight=None, batch_axis=0, **kwargs)
        self._alpha = alpha
        self._beta = beta
        self._top_k = top_k
        self._num_class = num_class
        self._num_in_class = num_in_class
        self._magrin = margin

    def _pair_distance(self, F, features):
        dot_product = F.dot(features, features.T)
        square_norm = F.sum(F.square(features), axis=1)
        distances = F.expand_dims(square_norm, 0) - 2.0 * dot_product + F.expand_dims(square_norm, 1)
        distances = F.maximum(distances, 0.0)
        mask = F.equal(distances, 0.0)
        distances = distances + mask * 1e-16
        distances = F.sqrt(distances)
        distances = distances * (1.0 - mask)
        return distances

    def _inter_class_loss(self, F, x, y):
        reshape_out = x.reshape((self._num_class, self._num_in_class, -1))
        centers = F.mean(reshape_out, axis=1)
        center_distance = self._pair_distance(F, centers)
        mask = F.array(
            1. - np.greater_equal.outer(np.arange(self._num_class), np.arange(self._num_class)).astype(np.float32))
        center_distance = center_distance * mask + (1. - mask) * 1e4
        center_distance = center_distance.reshape((-1,))
        inter_class_loss = F.maximum(self._magrin - F.min(center_distance), 0)
        return inter_class_loss

    def _intra_class_loss(self, F, x, y):
        intra_class_loss = F.array([0.])
        for i in range(self._num_class):
            same_label_feature = x[i * self._num_in_class:(i + 1) * self._num_in_class, :]
            same_label_distance = self._pair_distance(F, same_label_feature)
            mask = F.array(
                1. - np.greater_equal.outer(np.arange(self._num_in_class), np.arange(self._num_in_class)).astype(
                    np.float32))
            same_label_distance = same_label_distance * mask
            same_label_distance = same_label_distance.reshape((-1,))
            top_k_distance = F.topk(same_label_distance, k=self._top_k, ret_typ='value', is_ascend=False)
            harmonic_mean = self._top_k / F.sum(1 / (top_k_distance + 1e-8))
            intra_class_loss = intra_class_loss + harmonic_mean
        return intra_class_loss

    def hybrid_forward(self, F, x, y):
        inter_class_loss = self._inter_class_loss(F, x, y)
        intra_class_loss = self._intra_class_loss(F, x, y)
        range_loss = self._alpha * inter_class_loss + self._beta * intra_class_loss
        return range_loss


class MPSLoss(Loss):
    """Computes the MPS Loss from
    `"DocFace: Matching ID Document Photos to Selfies"
    <https://arxiv.org/abs/1805.02283>`_paper.

    Parameters
    ----------
    m: float
        Margin parameter for loss.

    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.
    """

    def __init__(self, m=1.0, **kwargs):
        super().__init__(weight=None, batch_axis=0, **kwargs)
        self.m = m

    @staticmethod
    def euclidean_distance(F, X, Y, sqrt=False):
        """Compute the distance between each X and Y.

        Args:
            X: a (m x d) tensor
            Y: a (d x n) tensor
            sqrt:

        Returns:
            diffs: an m x n distance matrix.
        """
        XX = F.sum(F.square(X), 1, keepdims=True)
        YY = F.sum(F.square(Y), 0, keepdims=True)
        XY = F.dot(X, Y)

        diffs = XX + YY - 2 * XY
        diffs = F.relu(diffs)
        if sqrt:
            diffs = F.sqrt(diffs)
        return diffs

    def hybrid_forward(self, F, pred1, pred2):
        pred1_norm = F.L2Normalization(pred1, mode="instance")
        pred2_norm = F.L2Normalization(pred2, mode="instance")

        # compute euclidean distance
        dist = -0.5 * self.euclidean_distance(F, pred1_norm, pred2_norm.transpose(), sqrt=False) + 1
        dist_pos = F.diag(dist)
        dist_neg = dist - F.diag(dist_pos)

        # get max dist between one image and others in a batch
        dist_neg_1 = F.expand_dims(F.max(dist_neg, axis=1), axis=1)
        dist_neg_2 = F.expand_dims(F.max(dist_neg, axis=0), axis=1)
        logits_neg = F.maximum(dist_neg_1, dist_neg_2)

        loss = (self.m + logits_neg - dist_pos) * 0.5
        return F.relu(loss)


class GitLoss(SoftmaxCrossEntropyLoss):
    """Computes the Git Loss from
    `"Git Loss for Deep Face Recognition"
    <https://arxiv.org/abs/1807.08512>`_paper.

    This implementation require the batch size not changing in training or validation.
    Commonly, it is ok, as when we train models last batch discard is applied, and no need
    for validation to compute the loss.

    Parameters
    ----------
    classes: int.
        Number of classes.
    embedding_size: int.
        Size of feature.
    lamda_c: float.
        The loss weight enforcing a trade-off between the softmax loss and center loss.
    lamda_g: float.
        The loss weight enforcing a trade-off between the softmax loss and git loss.
    batch_size_per_gpu: int.
        This size is sample numbers in each gpu or device, not total batch size
    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.

    """

    def __init__(self, classes, embedding_size, lamda_c, lamda_g, batch_size_per_gpu,
                 weight_initializer=init.Xavier(magnitude=2.24), dtype='float32', **kwargs):
        super().__init__(**kwargs)
        self._lamda_c = lamda_c
        self._lamda_g = lamda_g
        self._classes = classes
        self._dtype = dtype
        self.centers = self.params.get('centers', shape=(classes, embedding_size), init=weight_initializer,
                                       dtype=dtype, allow_deferred_init=True)
        self.mask = self.params.get_constant('mask', np.expand_dims(1 - np.eye(int(batch_size_per_gpu)), axis=2))

    def hybrid_forward(self, F, x, label, embeddings, centers, mask, sample_weight=None):
        centers_selected = F.take(centers, label)

        # Softmax
        loss_sm = super().hybrid_forward(F, x, label, sample_weight)
        onehot_label = F.one_hot(label, depth=self._classes, dtype=self._dtype)

        # loss center
        label_hist = F.sum(onehot_label, axis=0)
        centers_count = F.take(label_hist, label)
        loss_c = F.sum(F.square(embeddings - centers_selected), 1) / centers_count

        # loss git
        diffs = F.broadcast_sub(F.expand_dims(embeddings, axis=1), F.expand_dims(centers_selected, 0))
        diffs = F.broadcast_mul(diffs, mask)
        loss_g = F.mean(1 / (1 + F.sum(F.square(diffs), axis=2)), axis=1)

        return loss_sm + self._lamda_c * 0.5 * loss_c + self._lamda_g * loss_g


class COCOLoss(SoftmaxCrossEntropyLoss):
    """Computes the COCO Loss from
    `"Rethinking Feature Discrimination and Polymerization for Large-scale Recognition"
    <https://arxiv.org/abs/1710.00870>`_paper.

    This loss can be replaced by NormDense with Softmax, it is not recommended to use this.

    Parameters
    ----------
    classes: int.
        Number of classes.
    embedding_size: int.
        Size of feature.
    alpha: float.
        The scaling parameter, a hypersphere with small alpha
        will limit surface area for embedding features.

    Outputs:
        - **loss**: loss tensor with shape (batch_size,). Dimensions other than
          batch_axis are averaged out.

    """

    def __init__(self, classes, embedding_size, alpha,
                 weight_initializer=init.Xavier(magnitude=2.24), dtype='float32', **kwargs):
        super().__init__(**kwargs)
        self._alpha = alpha
        self._classes = classes
        self.centers = self.params.get('centers', shape=(classes, embedding_size), init=weight_initializer,
                                       dtype=dtype, allow_deferred_init=True)

    def hybrid_forward(self, F, embeddings, label, centers, sample_weight=None):
        norm_embs = self._alpha * F.L2Normalization(embeddings, mode='instance', name='fc1n')
        norm_centers = F.L2Normalization(centers, mode='instance', name='center_norm')
        outputs = F.dot(norm_embs, norm_centers, transpose_b=True)
        return super().hybrid_forward(F, outputs, label, sample_weight)


_losses = {
    'softmax': SoftmaxCrossEntropyLoss,
    'arcface': ArcLoss,
    'triplet': TripletLoss,
    'ringloss': RingLoss,
    'cosLoss': CosLoss,
    'l2softmax': L2Softmax,
    'asoftmax': ASoftmax,
    'centerloss': CenterLoss,
    'contrastiveloss': ContrastiveLoss,
    'lgmloss': LGMLoss,
    'mpsoss': MPSLoss,
    'gitloss': GitLoss,
    'cocoloss': COCOLoss,
}


def get_loss(name, **kwargs):
    """
    Parameters
    ----------
    name : str
        Name
    kwargs : str
        Params
    Returns
    -------
    HybridBlock
        The loss.
    """
    name = name.lower()
    if name not in _losses:
        err_str = '"%s" is not among the following losses list:\n\t' % (name)
        err_str += '%s' % ('\n\t'.join(sorted(_losses.keys())))
        raise ValueError(err_str)
    loss = _losses[name](**kwargs)
    return loss


def get_loss_list():
    """Get the entire list of loss names in losses.
    Returns
    -------
    list of str
        Entire list of loss names in losses.
    """
    return _losses.keys()



# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""This module provides popular face recognition datasets."""

from .dataset import *

"""
lfw:
`"Labeled Faces in the Wild: A Database for Studying Face Recognition in Unconstrained Environments"
 <http://vis-www.cs.umass.edu/lfw/lfw.pdf>_paper`

calfw:
`"A Database for Studying Cross-Age Face Recognition in Unconstrained Environments"
 <http://arxiv.org/abs/1708.08197>_paper`
 
cplfw:
`"Cross-pose LFW: A database for studying cross-pose face recognition in unconstrained environments"
 <http://www.whdeng.cn/CPLFW/Cross-Pose-LFW.pdf>_paper`

cfp_fp, cfp_ff:
`"Frontal to Profile Face Verification in the Wild"
 <http://www.cfpw.io/paper.pdf>_paper`

agedb_30:
`"AgeDB: the first manually collected, in-the-wild age database"
 <https://ibug.doc.ic.ac.uk/media/uploads/documents/agedb.pdf>_paper`

vgg2_fp:
`"VGGFace2: A dataset for recognising faces across pose and age"
 <https://arxiv.org/abs/1710.08092>_paper`

 
"""
datasets = {"lfw": FRValDataset,
            "calfw": FRValDataset,
            "cplfw": FRValDataset,
            "cfp_fp": FRValDataset,
            "agedb_30": FRValDataset,
            "cfp_ff": FRValDataset,
            "vgg2_fp": FRValDataset,
            "emore": FRTrainRecordDataset,
            "vgg": FRTrainRecordDataset,
            "webface": FRTrainRecordDataset,
            "asian_celeb": FRTrainRecordDataset,
            "deep_glint": FRTrainRecordDataset
            }


def get_recognition_dataset(name, **kwargs):
    return datasets[name](name, **kwargs)



# MIT License
#
# Copyright (c) 2018 Haoxintong
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
""""""
import mxnet as mx
import numpy as np
from sklearn.model_selection import KFold
from scipy import interpolate

__all__ = ["FaceVerification"]


class FaceVerification(mx.metric.EvalMetric):
    """ Compute confusion matrix of 1:1 problem in face verification or other fields.
    Use update() to collect the outputs and compute distance in each batch, then use get() to compute the
    confusion matrix and accuracy of the val dataset.

    Parameters
    ----------
    nfolds: int, default is 10

    thresholds: ndarray, default is None.
        Use np.arange to generate thresholds. If thresholds=None, np.arange(0, 2, 0.01) will be used for
        euclidean distance.

    far_target: float, default is 1e-3.
        This is used to get the verification accuracy of expected far.

    dist_type: int, default is 0.
        Option value is {0, 1}, 0 for euclidean distance, 1 for cosine similarity. Here for cosine distance,
        we use `1 - cosine` as the final distances.

    """

    def __init__(self, nfolds=10, thresholds=None, far_target=1e-3, dist_type=0):
        super().__init__("FaceVerification")
        self.far_target = far_target
        self._nfolds = nfolds
        self._dists = []
        self._issame = []
        default_thresholds = np.arange(0, 2, 0.01) if dist_type == 0 else np.arange(0, 1, 0.01)
        self._thresholds = thresholds if thresholds is not None else default_thresholds
        self.reset()
        self._dist_type = dist_type

    # noinspection PyMethodOverriding
    def update(self, labels: mx.nd.NDArray, embeddings0: mx.nd.NDArray, embeddings1: mx.nd.NDArray):
        """

        :param labels: NDArray.
        :param embeddings0: NDArray.
        :param embeddings1: NDArray.
        :return:
        """

        embeddings0 = embeddings0.asnumpy() if not isinstance(embeddings0, np.ndarray) else embeddings0
        embeddings1 = embeddings1.asnumpy() if not isinstance(embeddings1, np.ndarray) else embeddings1
        labels = labels.asnumpy() if not isinstance(labels, np.ndarray) else labels

        if self._dist_type == 0:
            diff = np.subtract(embeddings0, embeddings1)
            dists = np.sqrt(np.sum(np.square(diff), 1))
        else:
            dists = 1 - np.sum(np.multiply(embeddings0, embeddings1), axis=1) / \
                    (np.linalg.norm(embeddings0, axis=1) * np.linalg.norm(embeddings1, axis=1))

        self._dists += [d for d in dists]
        self._issame += [l for l in labels]

    def get(self):
        tpr, fpr, accuracy = calculate_roc(self._thresholds, np.asarray(self._dists),
                                           np.asarray(self._issame), self._nfolds)

        val, val_std, far = calculate_val(self._thresholds, np.asarray(self._dists),
                                          np.asarray(self._issame), self.far_target, self._nfolds)
        acc, acc_std = np.mean(accuracy), np.std(accuracy)
        return tpr, fpr, acc, val, val_std, far, acc_std

    def reset(self):
        self._dists = []
        self._issame = []


# code below is modified from project <Facenet (David Sandberg)>
class LFold:
    def __init__(self, n_splits=2, shuffle=False):
        self.n_splits = n_splits
        if self.n_splits > 1:
            self.k_fold = KFold(n_splits=n_splits, shuffle=shuffle)

    def split(self, indices):
        if self.n_splits > 1:
            return self.k_fold.split(indices)
        else:
            return [(indices, indices)]


def calculate_roc(thresholds, dist, actual_issame, nrof_folds=10):
    assert len(dist) == len(actual_issame), "Shape of predicts and labels mismatch!"

    nrof_pairs = len(dist)
    nrof_thresholds = len(thresholds)
    k_fold = LFold(n_splits=nrof_folds, shuffle=False)

    tprs = np.zeros((nrof_folds, nrof_thresholds))
    fprs = np.zeros((nrof_folds, nrof_thresholds))
    accuracy = np.zeros((nrof_folds,))
    indices = np.arange(nrof_pairs)
    dist = np.array(dist)

    for fold_idx, (train_set, test_set) in enumerate(k_fold.split(indices)):
        # Find the best threshold for the fold
        acc_train = np.zeros((nrof_thresholds,))
        for threshold_idx, threshold in enumerate(thresholds):
            _, _, acc_train[threshold_idx] = calculate_accuracy(threshold, dist[train_set], actual_issame[train_set])

        best_threshold_index = np.argmax(acc_train)
        for threshold_idx, threshold in enumerate(thresholds):
            tprs[fold_idx, threshold_idx], \
            fprs[fold_idx, threshold_idx], _ = calculate_accuracy(threshold, dist[test_set],
                                                                  actual_issame[test_set])
        _, _, accuracy[fold_idx] = calculate_accuracy(thresholds[best_threshold_index], dist[test_set],
                                                      actual_issame[test_set])

    tpr = np.mean(tprs, 0)
    fpr = np.mean(fprs, 0)
    return tpr, fpr, accuracy


def calculate_accuracy(threshold, dist, actual_issame):
    predict_issame = np.less(dist, threshold)
    tp = np.sum(np.logical_and(predict_issame, actual_issame))
    fp = np.sum(np.logical_and(predict_issame, np.logical_not(actual_issame)))
    tn = np.sum(np.logical_and(np.logical_not(predict_issame), np.logical_not(actual_issame)))
    fn = np.sum(np.logical_and(np.logical_not(predict_issame), actual_issame))

    tpr = 0 if (tp + fn == 0) else float(tp) / float(tp + fn)
    fpr = 0 if (fp + tn == 0) else float(fp) / float(fp + tn)
    acc = float(tp + tn) / dist.size
    return tpr, fpr, acc


def calculate_val(thresholds, dist, actual_issame, far_target, nrof_folds=10):
    assert len(dist) == len(actual_issame), "Shape of predicts and labels mismatch!"

    nrof_pairs = len(dist)
    nrof_thresholds = len(thresholds)
    k_fold = LFold(n_splits=nrof_folds, shuffle=False)

    val = np.zeros(nrof_folds)
    far = np.zeros(nrof_folds)
    indices = np.arange(nrof_pairs)
    dist = np.array(dist)

    for fold_idx, (train_set, test_set) in enumerate(k_fold.split(indices)):
        # Find the threshold that gives FAR = far_target
        far_train = np.zeros(nrof_thresholds)
        for threshold_idx, threshold in enumerate(thresholds):
            _, far_train[threshold_idx] = calculate_val_far(threshold, dist[train_set], actual_issame[train_set])

        if np.max(far_train) >= far_target:
            f = interpolate.interp1d(far_train, thresholds, kind='slinear')
            threshold = f(far_target)
        else:
            threshold = 0.0
        val[fold_idx], far[fold_idx] = calculate_val_far(threshold, dist[test_set], actual_issame[test_set])

    val_mean = np.mean(val)
    val_std = np.std(val)
    far_mean = np.mean(far)
    return val_mean, val_std, far_mean


def calculate_val_far(threshold, dist, actual_issame):
    predict_issame = np.less(dist, threshold)
    true_accept = np.sum(np.logical_and(predict_issame, actual_issame))
    false_accept = np.sum(np.logical_and(predict_issame, np.logical_not(actual_issame)))
    n_same = np.sum(actual_issame)
    n_diff = np.sum(np.logical_not(actual_issame))

    val = float(true_accept) / float(n_same)
    far = float(false_accept) / float(n_diff)
    return val, far



# @File  : model_zoo.py
# @Author: X.Yang
# @Contact : pistonyang@gmail.com
# @Date  : 19-2-19

from .mobile_facenet import *
from .attention_net import *
from .se_resnet import *

__all__ = ['get_model', 'get_model_list']

_models = {
    'l_se_resnet18v2': se_resnet18_v2,
    'l_se_resnet50v2': se_resnet50_v2,
    'l_se_resnet101v2': se_resnet101_v2,
    'mobilefacenet': get_mobile_facenet,
    'attentionfacenet56': attention_net56,
    'attentionfacenet92': attention_net92,
}


def get_model(name, **kwargs):
    """Returns a model by name
    Parameters
    ----------
    name : str
        Name of the model.
    classes : int
        Number of classes for the output layer.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    Returns
    -------
    HybridBlock
        The model.
    """
    name = name.lower()
    if name not in _models:
        err_str = '"%s" is not among the following model list:\n\t' % (name)
        err_str += '%s' % ('\n\t'.join(sorted(_models.keys())))
        raise ValueError(err_str)
    net = _models[name](**kwargs)
    return net


def get_model_list():
    """Get the entire list of model names in model_zoo.
    Returns
    -------
    list of str
        Entire list of model names in model_zoo.
    """
    return _models.keys()
